function make_callback(env, P)
  isperiodic = false
  cbfn_equation(u, t) = callback_equation(u, t, env, P, isperiodic, P.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P, isperiodic, P.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, isperiodic, P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  return callbackset
end


function make_callback(env, P::Project, bdryconds::BoundaryConditions)
  cbfn_equation(u, t) = callback_equation(u, t, env, P, isperiodic(bdryconds), P.equation)
  cb_equation         = FunctionCallback(cbfn_equation,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  cbfn_tci(u, t)      = callback_tci(u, t, env, P, isperiodic(bdryconds), P.tci)
  cb_tci              = FunctionCallback(cbfn_tci,
                                         CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, isperiodic(bdryconds), P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  return callbackset
end


function make_callback(env, P::Project2d)
  # cbfn_equation(u, t) = callback_equation(u, t, env, P, true, P.equation)
  # cb_equation         = FunctionCallback(cbfn_equation,
  #                                        CallbackTiming(every_iteration=1,every_dt=0))
  # cbfn_tci(u, t)      = callback_tci(u, t, env, P, true, P.tci)
  # cb_tci              = FunctionCallback(cbfn_tci,
  #                                        CallbackTiming(every_iteration=1, every_dt=0))
  cbfn_hrsc(u, t)     = callback_hrsc(u, t, env, P, true, P.hrsc, env.mesh)
  cb_hrsc             = FunctionCallback(cbfn_hrsc,
                                         CallbackTiming(every_iteration=1,every_dt=0))
  # callbackset = CallbackSet(cb_equation, cb_tci, cb_hrsc)
  callbackset = CallbackSet(cb_hrsc)
  return callbackset
end


#######################################################################
#                              Equation                               #
#######################################################################


function callback_equation(state_u, state_t, env, P, isperiodic, eq::Equation)
  @unpack cache = env
  wrap_dynamic_variables!(cache, state_u)
  @unpack E, Em1, flx_E, flx_Em1 = get_static_variables(cache)
  @unpack t, tm1 = get_global_variables(cache)
  tm1     .= t
  t       .= state_t
  Em1     .= E
  flx_Em1 .= flx_E
  broadcast_volume_2!(entropy_variables, P.equation, env.mesh)
end


#######################################################################
#                                HRSC                                 #
#######################################################################


callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Nothing, mesh) = nothing
callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::AbstractReconstruction, mesh) = nothing


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::Maybe{HRSC.AbstractHRSC}, mesh)
  TODO(callback_hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.AbstractArtificialViscosity, mesh)
  @unpack cache, mesh           = env
  @unpack equation              = P
  @unpack max_v, v              = get_static_variables(cache)
  @unpack mu, tmp_mu, cellmax_v = get_cell_variables(cache)
  @unpack D, S, tau             = get_dynamic_variables(cache)
  @unpack t, tm1                = get_global_variables(cache)

  max_v = cellmax_v
  t[1] <= 0.0 && tm1[1] <= 0 && return
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end

  # HRSC.compute_viscosity!(
  #     mu, tmp_mu,
  #     (D,S,tau), cellmax_v,
  #     hrsc)
  HRSC.compute_viscosity!(
      mu,
      D, cellmax_v,
      hrsc)

end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.TCIViscosity, mesh)
  @unpack cache, mesh = P
  @unpack mu, flag    = get_cell_variables(cache)
  HRSC.compute_viscosity!(mu, flag, hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh)
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache       = env
  @unpack mu          = get_cell_variables(cache)
  @unpack smoothed_mu = get_static_variables(cache)
  hrsc.smoother(smoothed_mu, mu, mesh, false)
end
function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.SmoothedArtificialViscosity, mesh::Mesh2d)
  error()
  callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc.av, mesh)
  @unpack cache       = env
  @unpack mu          = get_cell_variables(mesh.cache)
  @unpack smoothed_mu = get_static_variables(mesh.cache)
  hrsc.smoother(smoothed_mu, mu, mesh, true, true)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh1d)
  @unpack cache, mesh               = env
  @unpack D                         = get_dynamic_variables(cache)
  @unpack E, Em1, flx_E, flx_Em1, v = get_static_variables(cache)
  @unpack mu, cellmax_v             = get_cell_variables(cache)
  @unpack t, tm1                    = get_global_variables(cache)

  max_v = cellmax_v
  t[1] <= 0.0 && tm1[1] <= 0 && return
  Npts, K = layout(mesh)
  mat_max_v = reshape(view(v, :), (Npts, K))
  for k = 1:K
    max_v[k] = maximum(view(mat_max_v, :, k))
  end

  HRSC.compute_viscosity!(
      mu,
      D, max_v, E, Em1, flx_E, flx_Em1, t[1], tm1[1],
      hrsc)
end


function callback_hrsc(state_u, state_t, env, P, isperiodic, hrsc::HRSC.EntropyViscosity, mesh::Mesh2d)
  @unpack equation   = P
  @unpack E, Em1, flx_E_x, flx_E_y,
          flxm1_E_x, flxm1_E_y, max_v,
          dxdX, dydY = get_static_variables(mesh.cache)
  @unpack mu         = get_cell_variables(mesh.cache)
  @unpack t, tm1     = get_global_variables(mesh.cache)

  # backup previous values
  @. Em1   = E
  @. flxm1_E_x = flx_E_x
  @. flxm1_E_y = flx_E_y
  @. tm1   = t
  @. t     = state_t

  # compute new values
  broadcast_volume_2!(entropy_variables, equation, mesh)
  broadcast_volume_2!(maxspeed, equation, mesh)

  # TODO HRSC no longer provides an interface ...
  HRSC.compute_viscosity!(
      mu,
      E #= TODO This is unused =#,
      max_v, E, Em1, flx_E_x, flx_E_y, flxm1_E_x, flxm1_E_y, t[1], tm1[1], dxdX, dydY,
      hrsc)
end


#######################################################################
#                                 TCI                                 #
#######################################################################


function callback_tci(state_u, state_t, env, P, isperiodic, tci::Nothing) end


function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.Threshold)
  error("Threshold TCI must be combined with an indicator that is supposed to be zero for " *
        "for smooth solutions, but not for shocks. TODO Implement an indicator variable.")
end


function callback_tci(state_u, state_t, env, P, isperiodic, tci::TCI.EntropyProduction)
  @warn "Figure out entropy fluxes for special relativistic euler equation!"
  TODO()
  @unpack cache, mesh                = env
  @unpack E, Em1, flx_E, flx_Em1, EP = get_static_variables(cache)
  @unpack flag                       = get_cell_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs     = get_bdry_variables(cache)
  @unpack t, tm1                     = get_global_variables(cache)
  TCI.compute_indicator!(flag, EP, EPjump_lhs, EPjump_rhs,
                         E, Em1, flx_E, flx_Em1, t[1], tm1[1],
                         tci)
end


function callback_tci(state_u, state_t, env, P, isperiodic,
    tci::Union{TCI.Diffusion, TCI.Minmod, TCI.Threshold,
               TCI.ModalDecayAverage, TCI.ModalDecayHighest})

  @unpack cache, mesh = env
  @unpack D, S, tau   = get_dynamic_variables(cache)
  @unpack flag, D_flag, S_flag, tau_flag = get_cell_variables(cache)
  TCI.compute_indicator!(D_flag, D, tci)
  TCI.compute_indicator!(S_flag, S, tci)
  TCI.compute_indicator!(tau_flag, tau, tci)
  _, K = layout(mesh)
  for k = 1:K
    Df, Sf, tauf = D_flag[k], S_flag[k], tau_flag[k]
    flag[k] = max(Df, Sf, tauf)
    # if flag[k] > 0.0
    #   flag[k] = 1.0
    # end
  end
  # 'smooth' flag with neighbouring cells by simple averaging
  tmp_flag = D_flag # reuse memory
  for k = 1:K
    k_lhs, k_rhs = periodic_index(k-1, K), periodic_index(k+1, K)
    if !isperiodic && k == 1
      tmp_flag[k] = (flag[k] + flag[k_rhs]) / 2
    elseif !isperiodic && k == K
      tmp_flag[k] = (flag[k_lhs] + flag[k]) / 2
    else
      tmp_flag[k] = (flag[k_lhs] + flag[k] + flag[k_rhs]) / 3
    end
  end
  @. flag = tmp_flag
end
